@@ -82,6 +82,21 @@ class TestEndpointFeatures(TestEndpointBase):
         kwargs = self.http_session.send.call_args[1]
         self.assertEqual(kwargs['timeout'], timeout_override)
 
+    def test_make_request(self):
+        self.endpoint.make_request(self.op, request_dict(), self.request_signer)
+        prepared_request = self.http_session.send.call_args[0][0]
+        self.http_session.send.assert_called_with(
+            prepared_request, verify=True, stream=False,
+            proxies={}, timeout=DEFAULT_TIMEOUT, allow_redirects=True)
+
+    def test_make_request_with_no_redirects(self):
+        self.endpoint.make_request(self.op, request_dict(), self.request_signer,
+                                   allow_redirects=False)
+        prepared_request = self.http_session.send.call_args[0][0]
+        self.http_session.send.assert_called_with(
+            prepared_request, verify=True, stream=False,
+            proxies={}, timeout=DEFAULT_TIMEOUT, allow_redirects=False)
+
     def test_make_request_with_proxies(self):
         proxies = {'http': 'http://localhost:8888'}
         self.endpoint.proxies = proxies
@@ -89,7 +104,7 @@ class TestEndpointFeatures(TestEndpointBase):
         prepared_request = self.http_session.send.call_args[0][0]
         self.http_session.send.assert_called_with(
             prepared_request, verify=True, stream=False,
-            proxies=proxies, timeout=DEFAULT_TIMEOUT)
+            proxies=proxies, timeout=DEFAULT_TIMEOUT, allow_redirects=True)
 
     def test_make_request_with_no_auth(self):
         self.endpoint.auth = None
@@ -237,8 +252,9 @@ class TestEndpointCreator(unittest.TestCase):
     def test_altus_endpoint(self):
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -248,8 +264,9 @@ class TestEndpointCreator(unittest.TestCase):
     def test_cdp_endpoint(self):
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -260,8 +277,9 @@ class TestEndpointCreator(unittest.TestCase):
         self.cdp_service_model.endpoint_prefix = 'prefix'
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -269,12 +287,13 @@ class TestEndpointCreator(unittest.TestCase):
         self.assertEqual(endpoint.host,
                          'https://prefix.us-west-1.cdp.cloudera.com:443')
 
-    def test_creates_endpoint_with_configured_url(self):
+    def test_create_endpoint_with_configured_url(self):
         endpoint_url = 'https://endpoint.url'
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=endpoint_url,
+            explicit_endpoint_url=endpoint_url,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -284,9 +303,10 @@ class TestEndpointCreator(unittest.TestCase):
     def test_create_endpoint_with_customized_timeout(self):
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url='https://example.com',
+            explicit_endpoint_url='https://example.com',
             scoped_config={},
             timeout=123,
+            region='default',
             response_parser_factory=self.factory_patch,
             tls_verification=False,
             retry_handler=self.retry_handler)
@@ -296,9 +316,10 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://endpoint-config.url'
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -309,9 +330,10 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://endpoint-config.url'
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"cdp_endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -323,8 +345,9 @@ class TestEndpointCreator(unittest.TestCase):
         expected = endpoint_url % self.service_model.endpoint_name
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=endpoint_url,
+            explicit_endpoint_url=endpoint_url,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -337,8 +360,9 @@ class TestEndpointCreator(unittest.TestCase):
         expected = endpoint_url % self.cdp_service_model.endpoint_prefix
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=endpoint_url,
+            explicit_endpoint_url=endpoint_url,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -350,9 +374,10 @@ class TestEndpointCreator(unittest.TestCase):
         expected = endpoint_url % self.service_model.endpoint_name
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -365,9 +390,10 @@ class TestEndpointCreator(unittest.TestCase):
         expected = endpoint_url % self.cdp_service_model.endpoint_prefix
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"cdp_endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -378,8 +404,9 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://%s.%s.endpoint.url'
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=endpoint_url,
+            explicit_endpoint_url=endpoint_url,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -390,8 +417,9 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://%s.%s.endpoint.url'
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=endpoint_url,
+            explicit_endpoint_url=endpoint_url,
             scoped_config={},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -402,9 +430,10 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://%s.%s.endpoint-config.url'
         endpoint = self.creator.create_endpoint(
             self.service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -415,9 +444,10 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://%s.%s.endpoint-config.url'
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"cdp_endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -428,9 +458,10 @@ class TestEndpointCreator(unittest.TestCase):
         endpoint_url = 'https://foo-api.internal.url'
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"cdp_endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
@@ -443,11 +474,185 @@ class TestEndpointCreator(unittest.TestCase):
         expected = endpoint_url % self.cdp_service_model.endpoint_prefix
         endpoint = self.creator.create_endpoint(
             self.cdp_service_model,
-            endpoint_url=None,
+            explicit_endpoint_url=None,
             scoped_config={"cdp_endpoint_url": endpoint_url,
                            "config1": "value1"},
+            region='default',
             timeout=123,
             response_parser_factory=self.factory_patch,
             tls_verification=False,
             retry_handler=self.retry_handler)
         self.assertEqual(endpoint.host, expected)
+
+    def test_create_endpoint_with_explicit_region_us_west_1(self):
+        endpoint_url = 'https://iamapi.us-west-1.altus.cloudera.com:443'
+        endpoint = self.creator.create_endpoint(
+            self.service_model,
+            explicit_endpoint_url=None,
+            scoped_config={},
+            region='us-west-1',
+            timeout=123,
+            response_parser_factory=self.factory_patch,
+            tls_verification=False,
+            retry_handler=self.retry_handler)
+        self.assertEqual(endpoint.host, endpoint_url)
+
+    def test_create_endpoint_with_explicit_region_eu_1(self):
+        endpoint_url = 'https://api.eu-1.cdp.cloudera.com:443'
+        endpoint = self.creator.create_endpoint(
+            self.service_model,
+            explicit_endpoint_url=None,
+            scoped_config={},
+            region='eu-1',
+            timeout=123,
+            response_parser_factory=self.factory_patch,
+            tls_verification=False,
+            retry_handler=self.retry_handler)
+        self.assertEqual(endpoint.host, endpoint_url)
+
+    def test_create_endpoint_with_region_from_config_us_west_1(self):
+        endpoint_url = 'https://api.us-west-1.cdp.cloudera.com:443'
+        endpoint = self.creator.create_endpoint(
+            self.cdp_service_model,
+            explicit_endpoint_url=None,
+            scoped_config={'cdp_region': 'us-west-1'},
+            region='default',
+            timeout=123,
+            response_parser_factory=self.factory_patch,
+            tls_verification=False,
+            retry_handler=self.retry_handler)
+        self.assertEqual(endpoint.host, endpoint_url)
+
+    def test_create_endpoint_with_region_from_config_eu_1(self):
+        endpoint_url = 'https://api.eu-1.cdp.cloudera.com:443'
+        endpoint = self.creator.create_endpoint(
+            self.cdp_service_model,
+            explicit_endpoint_url=None,
+            scoped_config={'cdp_region': 'eu-1'},
+            region='default',
+            timeout=123,
+            response_parser_factory=self.factory_patch,
+            tls_verification=False,
+            retry_handler=self.retry_handler)
+        self.assertEqual(endpoint.host, endpoint_url)
+
+
+class TestEndpointResolver(unittest.TestCase):
+    def test_resolve_login_endpoint(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={},
+            region=None,
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')
+
+    def test_resolve_login_endpoint_with_explicit_url(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url="http://explict.url/path?accountId=foo",
+            config={},
+            region=None,
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'http://explict.url/path?accountId=foo')
+
+    def test_resolve_login_endpoint_with_login_url_from_config(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={'login_url': 'http://config.url/path?accountId=foo'},
+            region='eu-1',  # region is be ignored when using 'login_url' from config
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'http://config.url/path?accountId=foo')
+
+    def test_resolve_login_endpoint_with_explicit_region_default(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={},
+            region='default',
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')
+
+    def test_resolve_login_endpoint_with_explicit_region_us_west_1(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={},
+            region='us-west-1',
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')
+
+    def test_resolve_login_endpoint_with_explicit_region_eu_1(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={},
+            region='eu-1',
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint,
+                         'https://console.eu-1.cdp.cloudera.com:443/consoleauth/login')
+
+    def test_resolve_login_endpoint_with_region_from_config_default(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={'cdp_region': 'default'},
+            region=None,
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')
+
+    def test_resolve_login_endpoint_with_region_from_config_us_west_1(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={'cdp_region': 'us-west-1'},
+            region=None,
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint, 'https://consoleauth.altus.cloudera.com:443/login')
+
+    def test_resolve_login_endpoint_with_region_from_config_eu_1(self):
+        endpoint_resolver = EndpointResolver()
+        endpoint = endpoint_resolver.resolve(
+            explicit_endpoint_url=None,
+            config={'cdp_region': 'eu-1'},
+            region=None,
+            service_name='LOGIN',
+            prefix=None,
+            products=['CDP'],
+            scheme='https',
+            port=443)
+        self.assertEqual(endpoint,
+                         'https://console.eu-1.cdp.cloudera.com:443/consoleauth/login')
